#    This is a part of the "Prototype Tagger", an unsupervised POS tagger developed
#    in the Hebrew University by Omri Abend, Roi Reichart and Ari Rappoport.
#    Copyright (C) 2010  Omri Abend, Roi Reichart and Ari Rappoport
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    If you have any questions, feel free to contact us in:
#    omria01@cs.huji.ac.il, roiri@cs.huji.ac.il
#
#    Please read the README file for more details.
#
#!/usr/local/bin/python
import re, sys, random
from math import log, sqrt, fabs
from parameters_dast import *
from average_linker import *

INFINITY = 1000000

# Utilities which are used by the prototype tagger algorithm.

# reads a list from a file where every line is an entry
def read_list_from_file(file_name):
    if file_name == "-":
        return []
    f = open(file_name)
    l = list()
    for line in f:
        l.append(line.strip())
    f.close()
    return l
# END OF METHOD

# Calculates the cosine metric between two vectors.
# if the flag of the morphology is turned on, it also takes morphological similarity into
# account.
def cos(v1, v2):
    dot_product = 0
    norm1 = 0
    norm2 = 0

    for key in v1.keys():
        dot_product += v1.get(key,0)*v2.get(key, 0)
        norm1 += v1[key]*v1[key]
    
    for key in v2.keys():
        norm2 += v2[key]*v2[key]

    dot_product = dot_product/sqrt(norm1*norm2)
    return dot_product
# END OF METHOD

# Adds two sparse vectors
def add_vectors(v1, v2):
    v = dict()
    for key in v1.keys():
        v[key] = v1[key]
    for key in v2.keys():
        v[key] = v1.get(key, 0) + v2[key]
    return v
# END OF METHOD

# Multiplies the vector by a scalar.
def multiply_by_scalar(v, scalar):
    return dict([(key, v[key] * scalar) for key in v.keys()])
# END OF METHOD

# Reads a dictionary from a Seginer style lexicon. Returns vector_didct a dictionary of the vectors and freq_dict which is a dictionary
# of the frequencies.
# f - the file handle for a lexicon
# number_of_entries_to_read - the number of entries which it should read.
# line is the last line of the previous iteration.
# min_freq is the minimal number of appearences which will be included in the dictionary.
def read_dictionary(f, number_of_entries_to_read = None, line = None, min_freq = -1):
    if number_of_entries_to_read == None:
        number_of_entries_to_read = INFINITY
    
    vector_dict = dict()
    freq_dict = dict()
    if line == None:
        line = f.readline()
    delimiter = re.compile("[\\s()]+")
    read_entries = 0
    
    while line and read_entries < number_of_entries_to_read:

        fields = delimiter.split(line.strip())
        word_name = fields[0]
        
        total_num = float(fields[1])
        if total_num < min_freq:
            break
        
        cur_vector = dict()
    
        try:
            block_name, temp_coordinates, line = read_lexicon_block(f, total_num)
            
            if block_name == None:
                break
            
            while temp_coordinates != None:
                if block_name == "Left 1:":
                    for coor in temp_coordinates.keys():
                        if coor[0] == "W": #if it's a word label
                            new_coor = 'L' + coor[1:]
                        else:
                            new_coor = 'l' + coor[1:]
                        cur_vector[new_coor] = temp_coordinates[coor]
    
                if block_name == "Right 1:":
                    for coor in temp_coordinates.keys():
                        if coor[0] == "W": #if it's a word label
                            new_coor = 'R' + coor[1:]
                        else:
                            new_coor = 'r' + coor[1:]
                        cur_vector[new_coor] = temp_coordinates[coor]
                
                block_name, temp_coordinates, line = read_lexicon_block(f, total_num, line)
    
            read_entries += 1
            vector_dict[word_name] = cur_vector
            freq_dict[word_name] = total_num
            line = block_name
        except ValueError:
            sys.stderr.write("Error in dictionary entry. Omitting."+ word_name + "\n")

    print 'Total of '+str(len(vector_dict))+' in dictionary.'
    return vector_dict, freq_dict, line
# END OF METHOD

# Reads a lexicon. 
# Reads a block from the lexicon. Returns its block name and the list of extracted
# entries.
# corresponding values. If the file does not start with a block, it returns the
# last line to be read.
# f - a handle to a file that starts with a block
# total_num - the total number of instances of the word this block belongs to
# entry - the dict it should be read to
# first_line - the first line of the block. If it should be read, you should input None
def read_lexicon_block(f, total_num, line = None):
    delimiter = re.compile("[\\s()]+")
    
    if line == None:
        line = f.readline()
        
    if (not line.startswith("   ")):
        return line, None, None
    
    block_name = line.strip()
    entry = dict()
    
    while line.strip() != "Seen:":
        line = f.readline()
        if len(line) == 0:
            return None, None, None
        if line.startswith(" \""): #an empty block...
            return block_name, entry, line
        if "Blocked" in line:
            fields = delimiter.split(line.strip())
            index = fields.index("Blocked:")
            entry["WSTOP"] = float(fields[index+1]) / total_num
            
    line = f.readline()

    while line.startswith("       "):
        fields = delimiter.split(line)
        #print fields
        i = 0
        while (i < len(fields)-1):
            if fields[i] == "":
                i = i + 1
                continue
            name = fields[i]
            num = float(fields[i+1])
            i = i + 2
            if '.' not in name:
                entry["W" + name] = num / total_num
            else:
                name = name.replace('.','')
                entry["L" + name] = num / total_num
        line = f.readline()

    return block_name, entry, line

# END OF METHOD

# Receives a dict that maps items of some sort to a numerical value.
# It returns a list of the form (value, key) sorted by value.
def get_sorted_list(d):
    items = [(v, k) for k, v in d.items()]
    items.sort()
    items.reverse()
    return items
# END OF METHOD

# Performs an average link clustering.
# cluster_lists - initial clustering.
def average_link_clustering(vect_dict, thresh, cluster_lists = None, freq_dict = None):
    if cluster_lists == None:
        avg_cos = average_cos(thresh, len(vect_dict.keys()))
        avg_cos.load_matrix_from_dict(vect_dict)
    else:
        avg_cos = average_cos(thresh, len(cluster_lists.keys()))
        avg_cos.load_matrix_from_dict_with_initial_clustering(vect_dict, cluster_lists)
    
    if freq_dict != None:
        avg_cos.load_freq_dict(freq_dict)
    
    cluster_lists, temp, number_of_iterations = avg_cos.avg_link_clustering()
    return cluster_lists, number_of_iterations
# END OF METHOD

# Calculates the cosine metric between two clusters.

def cos_clusters(cluster1, cluster2, d):
    average = 0.0
    for v1 in cluster1:
        for v2 in cluster2:
            average += cos(d[v1], d[v2])
    return average/(len(cluster1)*len(cluster2))
# END OF METHOD

# Returns the cosine dictionary when there is a given clustering.
# cluster_lists - the given clustering.
# vect_dict - the vector dictionary
def get_cos_dict_clusters(cluster_lists, vect_dict):
    cos_dict = dict()
    for cluster_name1 in cluster_lists.keys():
        for cluster_name2 in cluster_lists.keys():
            if cluster_name1 == cluster_name2:
                continue
            pair = (cluster_name1, cluster_name2)
            cos_dict[pair] = cos_clusters(cluster_lists[cluster_name1], 
                                          cluster_lists[cluster_name2], vect_dict)
    return cos_dict
# END OF METHOD

# Returns the maximal pair in this dict which maps unordered pairs to a value.
# A pair of the maximum set and the maximum value will be returned.
def arg_max(d):
    return max(d, key = d.get)
# END OF METHOD

def arg_min(d):
    return min(d, key = d.get)
# END OF METHOD

# The method receives a dictionary of lists. It assumes these lists are disjoint.
# It returns the key of the entry in which the given object is found.
# If the key is not found, it returns None
# cluster_lists - the dictionary 
# word - the object
def in_what_entry(cluster_lists, word):
    for k in cluster_lists.keys():
        if word in cluster_lists[k]:
            return k
    return None
# END OF METHOD 

# Given a dictionary that maps a cluster name to a cluster list, it returns a dictionary 
# which maps a word to its cluster.
def get_reversed_dictionary(lists_dict):
    output = dict()
    for key in lists_dict.keys():
        for word in lists_dict[key]:
            output[word] = key
    return output  
# END OF METHOD

# We use the clusters to update the vectors. That is, if two words had the same
# cluster, there values are added.
# vect_dict - the dictionary
# clusters_lists - the lists of the clusters. Each word may belong to at most one cluster
def update_dictionary_by_clusters(vect_dict, clusters_lists):
    if clusters_lists == None or len(clusters_lists) == 0:
        return vect_dict
    
    in_what_entry = get_reversed_dictionary(clusters_lists)
    for word in vect_dict.keys():
        cur_vector = vect_dict[word]
        check_sum = sum_vector(cur_vector)

        visited = []
        coordinates = [coor for coor in cur_vector.keys()]
        for coor1 in coordinates:
            if coor1 in visited:
                continue
            visited.append(coor1)
            for coor2 in coordinates:
                if coor2 in visited:
                    continue
                cluster1 = in_what_entry.get(coor1[1:], None)
                cluster2 = in_what_entry.get(coor2[1:], None)
                
                if (cluster1 != None and cluster1 == cluster2 and coor1[0] == coor2[0]):
                    visited.append(coor2)
                    new_val = cur_vector[coor1] + cur_vector[coor2] 
                    
                    del cur_vector[coor2]
                    cur_vector[coor1] = new_val
        
        # Changing the names of the coordinates
        coordinates = [coor for coor in cur_vector.keys()]
        for coor in coordinates:
            cluster = in_what_entry.get(coor[1:], None)
            if cluster != None:
                new_coor = coor[0] + cluster
                val = cur_vector[coor]
                del cur_vector[coor]
                cur_vector[new_coor] = val

        # Updating d
        vect_dict[word] = cur_vector
        
    return vect_dict
# END OF METHOD

# Summing the dictionary over all the entries.
def sum_vector(vect_dict):
    S = 0
    for k in vect_dict.keys():
        S = S + vect_dict[k]
    return S
# END OF METHOD

# returns a dict which maps each cluster name to its frequency
def calc_cluster_frequency(clusters_lists, freq_dict):
    cluster_freq = dict()
    for cluster in clusters_lists.keys():
        cur_freq = 0
        l = clusters_lists[cluster]
        for word in l:
            cur_freq += freq_dict[word]
        cluster_freq[cluster] = cur_freq
    
    return cluster_freq
# END OF METHOD

# Finds the prominent clusters.
# clusters_lists - a dict of the clusters. Maps a cluster name to the list of words in it.
# freq_dict - a dictionary which maps a word to its frequency of occurrence
# vect_dict - a dictionary which maps a word to its context vector
# cos_dict - a dictionary which maps every pair of clusters to its cosine
# N - the number of outputted cluster
def prominent_clusters(clusters_lists, freq_dict, vect_dict, cos_dict, N):
    prominent_clusters = []
    
    if clusters_lists == None or len(clusters_lists) == 0:
        return prominent_clusters
    
    cluster_freq = calc_cluster_frequency(clusters_lists, freq_dict)
    sorted_cluster_freq = [v[1] for v in get_sorted_list(cluster_freq)]
    
    prominent_clusters.append(sorted_cluster_freq[0])
    
    while len(prominent_clusters) < N and len(prominent_clusters) < len(clusters_lists.keys()):
        new_prominent = get_farthest_cluster(clusters_lists, prominent_clusters, cos_dict)
        prominent_clusters.append(new_prominent)
        
    return prominent_clusters
# END OF METHOD

# Returns the cluster which is farthest away from all the clusters in the given list.
# That is, the cluster whose minimal cosine from all the other clusters is minimal
def get_farthest_cluster(clusters_lists, source_clusters, cos_clusters):
    max_cos_dict = dict()
    
    for cluster in clusters_lists.keys():
        if not cluster in source_clusters:
            max_cos = -10 #-infinity
            for source_cluster in source_clusters:
                cur_cos = cos_clusters.get((source_cluster, cluster))
                if cur_cos == None:
                    cur_cos = cos_clusters[(cluster, source_cluster)]
                if max_cos < cur_cos:
                    max_cos = cur_cos
            max_cos_dict[cluster] = max_cos
    return arg_min(max_cos_dict)
# END OF METHOD

# Returns the prototype number which contains the closest word by average.
# Returns the class of the prototype which is closest on average to the words in category.
def get_closest_prototype_cluster(category, prototypes, vector_dict):
    word_vectors = [vector_dict[v] for v in category]
    protos_avg_cosines = dict()
    class_index = 0
    for prototype_class in prototypes:
        for prototype_member in prototype_class:
            cur_avg = 0
            proto_vector = vector_dict[prototype_member]
            for vector in word_vectors:
                cur_avg += cos(vector, proto_vector)
            cur_avg = cur_avg / len(category)
            protos_avg_cosines[(prototype_member, class_index)] = cur_avg
        class_index += 1
    closest_prototype = arg_max(protos_avg_cosines)
    return closest_prototype[1]
# END OF METHOD

# Remove the classes under a certain size.
# clusters_list - a dictionary that maps a cluster name to a list of members of that class.
# size_thresh - the minimal allowable size
def remove_small_clusters(clusters_list, size_thresh):
    if clusters_list == None:
        return
    for cluster_name in clusters_list.keys():
        items = clusters_list[cluster_name]
        if len(items) < size_thresh:
            del clusters_list[cluster_name]
# END OF METHOD

# Returns a list of several lists, splitted according to the max_list_size
def split_list(L, max_list_size):
    if max_list_size <= 0:
        return [L]
    base_index = 0
    output = []
    while base_index < len(L):
        cur_sublist = L[base_index:base_index+max_list_size]
        base_index += max_list_size
        output.append(cur_sublist)
    return output
# END OF METHOD

